from typing import Tuple, Union
from abc import ABC

import numpy as np
from scipy import stats
from scipy.optimize import minimize

# pymc is optional - just be silent on failed import
try:
    import pymc as pm
except:
    pass


class BasePrior(ABC):
    def __init__(self, name: str, size: int = 1):
        self.name = name
        self.size = size

    def bounds(self, ci=1.0) -> Tuple[float, float]:
        return self.rv.interval(ci)

    def ppf(self, q):
        """Probability Percentage Function at q
        Defaults to using the underlying scipy distribution function

        Args:
            q: Quantile (float or arraylike) at which to evaluate ppf

        Returns:
            typeof(q): The ppf values
        """

        return self.rv.ppf(q)

    def logpdf(self, x):
        """Log Probability Density Function at x
        Defaults to using the underlying scipy distribution function

        Args:
            x: Value (float or arraylike) at which to evaluate logpdf

        Returns:
            typeof(x): The logpdf values
        """
        return self.rv.logpdf(x)

    def __repr__(self):
        return f"{type(self)} {self.name}"


# Union type for hierarchical/dispersion parameters
DistriParam = Union[float, BasePrior]


class BetaPrior(BasePrior):
    """
    A beta distributed prior.
    """

    def __init__(self, name: str, mean: float, ci: Tuple[float, float], size=1):
        super().__init__(name, size)
        self.mean = mean
        self.ci = ci
        self.name = name
        self._find_distri_params()
        self.rv = stats.beta(**self.distri_params)

    def bounds(self, ci=1.0) -> Tuple[float, float]:
        ret = self.rv.interval(ci)
        if isinstance(ci, float):
            return ret[0][0], ret[1][0]
        else:
            return ret

    def _find_distri_params(self, ci_width=0.95):
        ci = self.ci
        mean = self.mean

        assert len(ci) == 2 and ci[1] > ci[0] and 0.0 < ci_width < 1.0
        percentile_low = (1.0 - ci_width) / 2.0
        percentile_up = 1.0 - percentile_low
        assert 0.0 < ci[0] < 1.0 and 0.0 < ci[1] < 1.0 and 0.0 < mean < 1.0

        def distance_to_minimise(a):
            b = a * (1.0 - mean) / mean
            vals = stats.beta.ppf([percentile_low, percentile_up], a, b)
            dist = sum([(ci[i] - vals[i]) ** 2 for i in range(2)])
            return dist

        sol = minimize(distance_to_minimise, [1.0], bounds=[(0.0, None)], tol=1.0e-32)
        best_a = sol.x
        best_b = best_a * (1.0 - mean) / mean
        self.distri_params = {"a": best_a, "b": best_b}


class UniformPrior(BasePrior):
    """
    A uniformily distributed prior.
    """

    def __init__(self, name: str, domain: Tuple[float, float], size=1):
        super().__init__(name)
        self.start, self.end = domain
        self.distri_params = {"loc": self.start, "scale": self.end - self.start}
        self.rv = stats.uniform(**self.distri_params)
        self.size = size

    def to_pymc(self):
        if self.size > 1:
            lower = np.repeat(self.start, self.size)
            upper = np.repeat(self.end, self.size)
        else:
            lower, upper = self.start, self.end
        return pm.Uniform(self.name, lower=lower, upper=upper)

    def __repr__(self):
        return f"UniformPrior bounds: {self.bounds()}"


class TruncNormalPrior(BasePrior):
    """
    A prior with a truncated normal distribution.
    """

    def __init__(
        self, name: str, mean: float, stdev: float, trunc_range: Tuple[float, float], size=1
    ):
        super().__init__(name, size)
        self.mean, self.stdev = mean, stdev
        self.trunc_range = tuple(trunc_range)
        self.distri_params = {
            "loc": self.mean,
            "scale": self.stdev,
            "a": (trunc_range[0] - mean) / stdev,
            "b": (trunc_range[1] - mean) / stdev,
        }
        self.rv = stats.truncnorm(**self.distri_params)

    def to_pymc(self):
        lower, upper = self.trunc_range
        if self.size > 1:
            lower = np.repeat(lower, self.size)
            upper = np.repeat(upper, self.size)
        return pm.TruncatedNormal(
            self.name, mu=self.mean, sigma=self.stdev, lower=lower, upper=upper
        )

    def __repr__(self):
        return f"TruncNormalPrior mean: {self.mean}, stdev: {self.stdev}, bounds: {self.bounds()}"



import numpy as np
import pymc as pm
import pytensor.tensor as pt

from estival.model import BayesianCompartmentalModel

import cloudpickle


def get_wrapped_ll(bcm: BayesianCompartmentalModel):
    """_summary_

    Args:
        bcm: The model to wrap

    Returns:
        A wrapped pytensor op for use in pymc
    """

    prior_types = []
    for k, prior in bcm.priors.items():
        if prior.size > 1:
            prior_types.append(pt.dvector)
        else:
            prior_types.append(pt.dscalar)

    # define a pytensor Op for our likelihood function
    class BCMLogLike(pt.Op):
        """
        Specify what type of object will be passed and returned to the Op when it is
        called. In our case we will be passing it a vector of values (the parameters
        that define our model) and returning a single "scalar" value (the
        log-likelihood)
        """

        itypes = prior_types

        # [pt.dscalar] * len(
        #    bcm.priors
        # )  # [pt.dvector]  # expects a vector of parameter values when called
        otypes = [pt.dscalar]  # outputs a single scalar value (the log likelihood)

        def __init__(self):
            """
            Initialise the Op with various things that our log-likelihood function
            requires. Below are the things that are needed in this particular
            example.

            Parameters
            ----------
            bcm: BayesianCompartmentalModel
            """

            # Capture the BayesianCompartmentalModel
            self.bcm = bcm

        def perform(self, node, inputs, outputs):
            params = inputs
            kwargs = {k: params[i] for i, k in enumerate(self.bcm.priors)}

            # call the log-likelihood function
            logl = self.bcm.loglikelihood(**kwargs)

            outputs[0][0] = np.array(logl)  # output the log-likelihood

    return BCMLogLike


def use_model(bcm: BayesianCompartmentalModel, include_ll=False) -> list:
    """Use a given BayesianCompartmentalModel for pymc sampling
    This should be called inside a model context like so

    with pm.Model():
        variables = use_model(bcm)
        pm.sample(step=[pm.DEMetropolis(variables)])

    Args:
        bcm: The BCM to use for sampling
        include_ll: Include loglikelihood in the sample outputs

    Returns:
        The list of variables to be passed to a sampler step
    """
    logl = get_wrapped_ll(bcm)()

    pymc_priors = []

    for k, prior in bcm.priors.items():
        pymc_priors.append(prior.to_pymc())

    invars = [pt.as_tensor_variable(v) for v in pymc_priors]

    # use a Potential to "call" the Op and include it in the logp computation
    ll = logl(*invars)
    pot = pm.Potential("loglikelihood", ll)

    if include_ll:
        pm.Deterministic("loglike", ll)

    return pymc_priors


def run_map(bcm_pkl, ival):
    bcm = cloudpickle.loads(bcm_pkl)
    with pm.Model() as model:
        variables = use_model(bcm, include_ll=False)
        map_est = pm.find_MAP(ival, include_transformed=False, progressbar=False)
    return map_est



from typing import Tuple, Dict
from multiprocessing import cpu_count, Pool

import cloudpickle
import pandas as pd
from arviz import InferenceData
import numpy as np

from estival.model import BayesianCompartmentalModel
from estival.utils.parallel import map_parallel

SampleIndex = Tuple[int, int]
ParamDict = Dict[str, float]


def likelihood_extras_for_idata(
    idata: InferenceData, bcm: BayesianCompartmentalModel, n_workers: int = None
) -> pd.DataFrame:
    """Calculate the likelihood extras (ll,lprior,lpost + per-target) for all
    samples in supplied InferenceData, returning a DataFrame.

    Note - input InferenceData must be the full (unburnt) idata

    Args:
        idata: The InferenceData to sample
        bcm: The BayesianCompartmentalModel (must be the same BCM used to generate idata)
        n_workers: Number of multiprocessing workers to use; defaults to cpu_count/2

    Returns:
        A DataFrame with index (chain, draw) and columns being the keys in ResultsData.extras
            - Use df.reset_index(level="chain").pivot(columns="chain") to move chain into column multiindex
    """
    n_workers = n_workers or int(cpu_count() / 2)

    accepted_df = idata.sample_stats.accepted.to_dataframe()

    # Get indices and samples of accepted runs only (ie unique valid paramsets)
    accepted_indices = [
        (chain, draw) for (chain, draw), accepted in accepted_df.iterrows() if accepted["accepted"]
    ]
    # accepted_samples_df = idata.posterior.to_dataframe().loc[accepted_indices]

    accept_mask = idata.sample_stats.accepted.data
    posterior_t = idata.posterior.transpose("chain", "draw", ...)

    components = {}
    for dv in posterior_t.data_vars:
        components[dv] = posterior_t[dv].data[accept_mask]

    accepted_si = SampleIterator(components, index=accepted_indices)
    # Get the likelihood extras for all accepted samples - this spins up a multiprocessing pool
    # pres = sample_likelihood_extras_mp(bcm, accepted_samples_df, n_workers)

    extras_df = likelihood_extras_for_samples(accepted_si, bcm, n_workers)

    # Create a DataFrame with the full index of the idata
    # This has a lot of redundant information, but it's still only a few Mb and
    # makes lookup _so_ much easier...
    filled_edf = pd.DataFrame(index=accepted_df.index, columns=extras_df.columns)

    for (chain, draw), accepted_s in accepted_df.iterrows():
        # Extract the bool from the Series
        accepted = accepted_s["accepted"]
        # Update the index if this sample is accepted - otherwise we'll
        # store the previous known good sample (ala MCMC)
        if accepted:
            last_good_sample_idx = (chain, draw)
        filled_edf.loc[(chain, draw)] = extras_df.loc[last_good_sample_idx]

    return filled_edf


def likelihood_extras_for_samples(
    samples_df: pd.DataFrame, bcm: BayesianCompartmentalModel, n_workers: int = None
) -> pd.DataFrame:
    def get_sample_extras(sample_params: Tuple[SampleIndex, ParamDict]) -> Tuple[SampleIndex, Dict]:
        """Run the BCM for a given set of parameters, and return its extras dictionary
        (likelihood, posterior etc)

        Args:
            sample_params: The parameter set to sample (indexed by chain,draw)

        Returns:
            A tuple of SampleIndex and the ResultsData.extras dictionary
        """

        idx, params = sample_params
        res = bcm.run(params, include_extras=True)
        return idx, res.extras

    pres = map_parallel(get_sample_extras, samples_df.iterrows(), n_workers)

    extras_dict = {
        "logposterior": {},
        "logprior": {},
        "loglikelihood": {},
    }

    base_fields = list(extras_dict)

    for idx, res in pres:
        for field in base_fields:
            extras_dict[field][idx] = float(res[field])
        for k, v in res["ll_components"].items():
            extras_dict.setdefault("ll_" + k, {})
            extras_dict["ll_" + k][idx] = float(v)

    extras_df = pd.DataFrame(extras_dict)

    return extras_df


class SampleIterator:
    """A simple container storing dicts of arrays, providing a means to iterate over
    the array items (and returning a dict of the items at each index)
    Designed to be a drop-in replacement for pd.DataFrame.iterrows (but supporting multidimensional arrays)
    """

    def __init__(self, components: dict, index=None):
        self.components = components
        self._calc_component_length()
        if index is None:
            self.index = np.arange(self.clen)
        else:
            assert (
                idxlen := len(index)
            ) == self.clen, f"Index length {idxlen} not equal to component length {self.clen}"
            self.index = index

    def _calc_component_length(self):
        clen = None
        for k, cval in self.components.items():
            if clen is None:
                clen = len(cval)
            else:
                assert len(cval) == clen, f"Length mismatch for {k} ({len(cval)}), should be {clen}"
        self.clen = clen

    def __iter__(self):
        for i in range(self.clen):
            out = {}
            for k, v in self.components.items():
                out[k] = v[i]
            yield out

    def iterrows(self):
        for i in range(self.clen):
            out = {}
            for k, v in self.components.items():
                out[k] = v[i]
            yield self.index[i], out

    def __getitem__(self, idx):
        out = {}
        for k, v in self.components.items():
            out[k] = v[idx]
        return out



from __future__ import annotations

from typing import List, Tuple, Union
from abc import ABC, abstractmethod
from copy import copy

import pandas as pd
import numpy as np

from jax import jit, scipy as jsp, numpy as jnp

from summer2.utils import Epoch

from .priors import DistriParam, BasePrior


class BaseTarget(ABC):
    name: str
    data: pd.Series
    _data_attrs: List = []

    def __init__(
        self, name: str, data: pd.Series, weight: float = 1.0, time_weights: pd.Series = None
    ):
        # Make things easier for calibration by sanitizing the data here
        self.name = name
        self.data = data

        # Should do some validation on this - ie make sure indices match data
        # if time_weights is None:
        #    time_weights = pd.Series(index=data.index, data=np.repeat(1.0 / len(data), len(data)))

        self.time_weights = time_weights
        self.weight = weight

    def get_priors(self):
        return []

    def filtered(self, index: pd.Index) -> BaseTarget:
        out_target = copy(self)
        valid_idx = index.intersection(self.data.index)
        out_target.data = out_target.data[valid_idx]

        for da in self._data_attrs:
            data = getattr(out_target, da)
            setattr(out_target, da, data[valid_idx])

        if self.time_weights is not None:
            new_time_weights = out_target.time_weights[valid_idx]
            out_target.time_weights = new_time_weights / new_time_weights.sum()
        return out_target

    @abstractmethod
    def get_evaluator(self, model_times: pd.Index, epoch: Epoch) -> TargetEvaluator:
        raise NotImplementedError()


class TargetEvaluator(ABC):
    def __init__(self, target: BaseTarget, model_times: pd.Index, epoch: Epoch = None):
        if not isinstance(target.data.index, pd.DatetimeIndex):
            if epoch is not None:
                model_times = epoch.dti_to_index(model_times)
        else:
            if epoch is None:
                raise Exception("Target data expressed as datetime but no ref_date set for model")

        self.target = target.filtered(model_times)
        self.data = self.target.data.to_numpy()
        for da in target._data_attrs:
            data = getattr(self.target, da)
            setattr(self, da, data.to_numpy())
        self.index = np.array([model_times.get_loc(t) for t in self.target.data.index])
        if self.target.time_weights is not None:
            self.time_weights = self.target.time_weights.to_numpy()
        else:
            self.time_weights = None

    @abstractmethod
    def evaluate(self, modelled: np.array, parameters: dict) -> float:
        raise NotImplementedError()


class NegativeBinomialEvaluator(TargetEvaluator):
    def __init__(self, target: BaseTarget, model_times: pd.Index, epoch: Epoch):
        super().__init__(target, model_times, epoch)

    def evaluate(self, modelled: np.array, parameters: dict) -> float:
        if isinstance(self.target.dispersion_param, BasePrior):
            n = parameters[self.target.dispersion_param.name]
        else:
            n = self.target.dispersion_param

        # We use the parameterisation based on mean and variance and assume define var=mean**delta
        mu = modelled[self.index]
        # work out parameter p to match the distribution mean with the model output
        p = mu / (mu + n)
        # Attempt to minimize -inf showing up
        p = jnp.where(p < 1e-16, 1e-16, p)
        # ll = np.sum(stats.nbinom.logpmf(self.data, n, 1.0 - p) * self.time_weights)
        ll = jsp.stats.nbinom.logpmf(self.data, n, 1.0 - p)

        if self.time_weights is not None:
            ll = ll * self.time_weights
            return jnp.sum(ll) * self.target.weight
        else:
            return jnp.mean(ll) * self.target.weight


class NegativeBinomialTarget(BaseTarget):
    """
    A calibration target sampled from a negative binomial distribution
    """

    def __init__(
        self,
        name: str,
        data: pd.Series,
        dispersion_param: DistriParam,
        weight: float = 1.0,
        time_weights: pd.Series = None,
    ):
        super().__init__(name, data, weight, time_weights)
        self.dispersion_param = dispersion_param

    def get_priors(self):
        if isinstance(self.dispersion_param, BasePrior):
            return [self.dispersion_param]
        else:
            return []

    def get_evaluator(self, model_times: pd.Index, epoch: Epoch) -> TargetEvaluator:
        return NegativeBinomialEvaluator(self, model_times, epoch)


class BinomialTarget(BaseTarget):
    """
    A calibration target sampled from a binomial distribution
    """

    def __init__(
        self,
        name: str,
        data: pd.Series,
        sample_sizes: pd.Series,
        weight: float = 1.0,
        time_weights: pd.Series = None,
    ):
        # Enforce floats for data and sample_sizes so TFP doesn't complain later
        super().__init__(name, data.astype(float), weight, time_weights)
        self._data_attrs = ["sample_sizes"]
        self.sample_sizes = sample_sizes.astype(float)

    def get_evaluator(self, model_times: pd.Index, epoch: Epoch) -> TargetEvaluator:
        return BinomialEvaluator(self, model_times, epoch)


class BinomialEvaluator(TargetEvaluator):
    def __init__(self, target: BaseTarget, model_times: pd.Index, epoch: Epoch):
        super().__init__(target, model_times, epoch)
        # Enforce this here so TFP-jax picks the right output types
        self.sample_sizes = self.sample_sizes.astype(float)

    def evaluate(self, modelled: np.array, parameters: dict) -> float:
        from tensorflow_probability.substrates import jax as tfp

        # use a binomial (n, p) where n is the sample size observed in the data and p the modelled proportion
        # We then evaluate the binomial density for k, which represents the numerator observed in the data
        n = self.target.sample_sizes
        p = modelled[self.index]
        k = self.target.data * n

        bdist = tfp.distributions.Binomial(total_count=n, probs=p)
        ll = bdist.log_prob(k)

        if self.time_weights is not None:
            ll = ll * self.time_weights
            return jnp.sum(ll) * self.target.weight
        else:
            return jnp.mean(ll) * self.target.weight


class TruncatedNormalTarget(BaseTarget):
    """
    A calibration target sampled from a truncated normal distribution
    """

    def __init__(
        self,
        name: str,
        data: pd.Series,
        trunc_range: Tuple[float, float],
        stdev: DistriParam,
        weight: float = 1.0,
        time_weights: pd.Series = None,
    ):
        super().__init__(name, data, weight, time_weights)
        self.trunc_range = trunc_range
        self.stdev = stdev

    def get_priors(self):
        if isinstance(self.stdev, BasePrior):
            return [self.stdev]
        else:
            return []

    def get_evaluator(self, model_times: pd.Index, epoch: Epoch) -> TargetEvaluator:
        return TruncatedNormalTargetEvaluator(self, model_times, epoch)


class TruncatedNormalTargetEvaluator(TargetEvaluator):
    def __init__(self, target: TruncatedNormalTarget, model_times: pd.Index, epoch: Epoch):
        super().__init__(target, model_times, epoch)

    def evaluate(self, modelled: np.array, parameters: dict) -> float:
        if isinstance(self.target.stdev, BasePrior):
            sd = parameters[self.target.stdev.name]
        else:
            sd = self.target.stdev

        from tensorflow_probability.substrates.jax import distributions as tfpd

        tdist = tfpd.TruncatedNormal(
            loc=self.data, scale=sd, low=self.target.trunc_range[0], high=self.target.trunc_range[1]
        )

        distri_params = {
            "scale": sd,
            "a": (self.target.trunc_range[0] - self.data) / sd,
            "b": (self.target.trunc_range[1] - self.data) / sd,
        }

        # ll = jsp.stats.truncnorm.logpdf(modelled[self.index], loc=self.data, **distri_params)

        ll = tdist.log_prob(modelled[self.index])

        if self.time_weights is not None:
            ll = ll * self.time_weights
            return jnp.sum(ll) * self.target.weight
        else:
            return jnp.mean(ll) * self.target.weight


class NormalTarget(BaseTarget):
    """
    A calibration target sampled from a normal distribution
    """

    def __init__(
        self,
        name: str,
        data: pd.Series,
        stdev: DistriParam,
        weight: float = 1.0,
        time_weights: pd.Series = None,
    ):
        super().__init__(name, data, weight, time_weights)
        self.stdev = stdev

    def get_priors(self):
        if isinstance(self.stdev, BasePrior):
            return [self.stdev]
        else:
            return []

    def get_evaluator(self, model_times: pd.Index, epoch: Epoch) -> TargetEvaluator:
        return NormalTargetEvaluator(self, model_times, epoch)


class NormalTargetEvaluator(TargetEvaluator):
    def __init__(self, target: BaseTarget, model_times: pd.Index, epoch: Epoch):
        super().__init__(target, model_times, epoch)

    def evaluate(self, modelled: np.array, parameters: dict) -> float:
        if isinstance(self.target.stdev, BasePrior):
            sd = parameters[self.target.stdev.name]
        else:
            sd = self.target.stdev

        ll = jsp.stats.norm.logpdf(modelled[self.index], loc=self.data, scale=sd)

        if self.time_weights is not None:
            ll = ll * self.time_weights
            return jnp.sum(ll) * self.target.weight
        else:
            return jnp.mean(ll) * self.target.weight


class CustomTargetEvaluator(TargetEvaluator):
    def __init__(self, target, model_times, eval_func, epoch):
        super().__init__(target, model_times, epoch)
        self._eval_func = eval_func

    def evaluate(self, modelled, parameters):
        return (
            self._eval_func(modelled[self.index], self.data, parameters, self.time_weights)
            * self.target.weight
        )


class CustomTarget(BaseTarget):
    def __init__(self, name, data, eval_func, weight=1.0, time_weights=None):
        """Build a Target that uses a custom evaluation function.
        Indexing and multiplication by weight factor is handled automatically,
        but time_weights are not (see below)

        For example:

        def least_squares(modelled, obs, parameters, time_weights):
            return (((modelled - obs) ** 2.0) * time_weights).sum()

        CustomTarget("example", example_data, least_squares)

        Args:
            name: Name (key) of output to evaluate against
            data: Series of data using same indexing conventions as model
            eval_func: Callable as described above
            weight (optional): Scales resulting output
            time_weights (optional): Passed to eval_func - Series with index matching data
        """
        super().__init__(name, data, weight, time_weights)
        self.eval_func = eval_func

    def get_evaluator(self, model_times, epoch: Epoch):
        return CustomTargetEvaluator(self, model_times, self.eval_func, epoch)


def get_target_sd(data: pd.Series) -> float:
    """Return a value such that the 95% CI of the associated normal distribution covers a width
       equivalent to 25% of the maximum value of the target.

    Args:
        data: The target data series

    Returns:
        Calculated sd
    """
    return 0.25 / 4.0 * max(data)



from typing import Callable, Iterable

from multiprocessing import Pool, cpu_count

import cloudpickle


# initialize a worker in the process pool
def process_init_cloudpickle(custom_cpkl):
    """Initialize a mp.Process with custom cloudpickled data

    Args:
        custom_cpkl: The cloudpickle.dumps output of the custom data
    """
    # declare global variable
    global custom_data
    # assign the global variable
    custom_data = cloudpickle.loads(custom_cpkl)


def generic_cpkl_worker(*args):
    """Worker function to be used with multiprocessing pools, where
    custom_data is a global initialized to a function, usually via
    process_init_cloudpickle

    Returns:
        Return type of the custom global function
    """
    global custom_data
    run_func = custom_data

    return run_func(*args)


def map_parallel(run_func: Callable, input_iterator: Iterable, n_workers: int = None):
    """Map the values of input_iterator over a function run_func, using n_workers parallel workers

    Args:
        run_func: The function to call over the mapped inputs
        input_iterator: An iterable containing the values to map
        n_workers: Number of processes used by Pool

    Returns:
        A list of values return by run_func
    """

    if n_workers is None:
        n_workers = int(cpu_count() / 2)

    with Pool(
        n_workers, initializer=process_init_cloudpickle, initargs=(cloudpickle.dumps(run_func),)
    ) as pool:
        pres = pool.map(generic_cpkl_worker, input_iterator)
        # +++ We seem to get intermittent hanging here, and the context manager _should_ handle cleanup...
        # pool.close()
        # pool.join()
    return pres



from typing import List
from dataclasses import dataclass

from summer2 import CompartmentalModel

from jax import jit
import numpy as np

import pandas as pd


@dataclass
class ResultsData:
    derived_outputs: pd.DataFrame
    extras: dict


class BayesianCompartmentalModel:
    def __init__(
        self,
        model: CompartmentalModel,
        parameters: dict,
        priors: list,
        targets: list,
        extra_ll=None,
    ):
        self.model = model

        self._model_parameters = model.get_input_parameters()

        self.parameters = parameters
        self.targets = {t.name: t for t in targets}

        for t in targets:
            priors = priors + t.get_priors()
        self.priors = {p.name: p for p in priors}

        self._ref_idx = self.model._get_ref_idx()
        if not isinstance(self._ref_idx, pd.Index):
            self._ref_idx = pd.Index(self._ref_idx)
        self.epoch = self.model.get_epoch()

        self._build_logll_funcs(extra_ll)

    def _build_logll_funcs(self, extra_ll=None):
        model_params = self.model.get_input_parameters()
        dyn_params = list(model_params.intersection(set(self.priors)))
        self.model.set_derived_outputs_whitelist(list(self.targets))

        self._ll_runner = self.model.get_runner(
            self.parameters, dyn_params, include_full_outputs=False
        )

        self.model.set_derived_outputs_whitelist([])
        self._full_runner = self.model.get_runner(
            self.parameters, dyn_params, include_full_outputs=False
        )

        self._evaluators = {}
        for k, t in self.targets.items():
            tev = t.get_evaluator(self._ref_idx, self.epoch)
            self._evaluators[k] = tev.evaluate

        @jit
        def logll(**kwargs):
            dict_args = capture_model_kwargs(self.model, **kwargs)
            res = self._ll_runner._run_func(dict_args)["derived_outputs"]

            logdens = 0.0
            for tk, te in self._evaluators.items():
                modelled = res[tk]
                logdens += te(modelled, kwargs)

            if extra_ll:
                logdens += extra_ll(kwargs)

            return logdens

        logll.__doc__ = f"""logll({', '.join([k for k in self.priors])})\n
        Run the model for a given set of parameters, and 
        return the loglikelihood of its outputs, including any values from extrall"""

        @jit
        def logll_multi(modelled_do, **kwargs):
            out_ll = {}

            for tk, te in self._evaluators.items():
                modelled = modelled_do[tk]
                out_ll[tk] = te(modelled, kwargs)

            if extra_ll:
                out_ll["extra_ll"] = extra_ll(kwargs)

            return out_ll

        self._logll_multi = logll_multi
        self.loglikelihood = logll

    def logprior(self, **parameters):
        lp = 0.0
        for k, p in self.priors.items():
            lp += np.sum(p.logpdf(parameters[k]))
        return lp

    def logposterior(self, **parameters):
        return self.loglikelihood(**parameters) + self.logprior(**parameters)

    def run(self, parameters: dict, include_extras=False) -> ResultsData:
        """Run the model for a given set of parameters.
        Note that only parameters specified as priors affect the outputs; other parameters
        are in-filled from the initial arguments supplied to BayesianCompartmentalModel

        Args:
            parameters: Dict of parameter key/values (as specified in priors)

        Returns:
            ResultsData, an extensible container with derived_outputs as a DataFrame
        """
        run_params = {k: v for k, v in parameters.items() if k in self._model_parameters}
        results = self._full_runner._run_func(run_params)

        if include_extras:
            extras = {}
            ll_components = self._logll_multi(results["derived_outputs"], **parameters)
            extras["ll_components"] = ll_components
            extras["loglikelihood"] = sum(ll_components.values())
            extras["logprior"] = self.logprior(**parameters)
            extras["logposterior"] = extras["logprior"] + extras["loglikelihood"]
        else:
            extras = {}

        return ResultsData(
            derived_outputs=pd.DataFrame(results["derived_outputs"], index=self._ref_idx),
            extras=extras,
        )

    def run_jax(self, parameters: dict) -> dict:
        """Run the jax run function for the model directly with the supplied parameters;
        meaning bcm.run_jax can be included in JIT calls

        Args:
            parameters: Dict of parameter key/values (as specified in priors)

        Returns:
            Results as per the summer2 jax runner
        """
        return self._full_runner._run_func(parameters)


def capture_model_kwargs(model: CompartmentalModel, **kwargs) -> dict:
    model_params = model.get_input_parameters()
    return {k: kwargs[k] for k in kwargs if k in model_params}
